@@ -22,6 +22,13 @@
 #define _GNU_SOURCE 1
 #endif
 
+#ifdef SUSE
+# define USE_PRELOAD
+# undef  CHECK_FORDEVPTS
+#else
+# undef  USE_PRELOAD
+# define CHECK_FORDEVPTS
+#endif
 #include <stdio.h>
 #include <termios.h>
 #include <sys/types.h>
@@ -29,8 +36,13 @@
 #include <sys/select.h>
 #include <sys/time.h>
 #include <sys/ioctl.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 #include <sys/sysinfo.h>
 #include <sys/stat.h>
+#ifdef CHECK_FORDEVPTS
+# include <sys/vfs.h>
+#endif
 #include <time.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -66,7 +78,8 @@ static struct makenode **nodevec;
 
 #define PBUF_SIZE	8192
 struct prg {
-  char *name;
+  const char *name;
+  const char *arg0;
   int num;
   int fd;
   pid_t pid;
@@ -81,6 +94,22 @@ static int inpar, par;
 static int pidpipe[2];
 static double iorate = 800.0;
 
+#ifdef USE_PRELOAD
+static enum { Unknown, Preload, NoPreload } ispreload = Unknown;
+
+static void sighandler_nopreload(int x __attribute__((unused)))
+{
+    ispreload = NoPreload;
+}
+
+static void sighandler_preload(int x __attribute__((unused)))
+{
+    ispreload = Preload;
+}
+
+# define SOCK_PATH "/dev/shm/preload_sock"
+#endif
+
 void *xcalloc(size_t nmemb, size_t size)
 {
   void *r;
@@ -150,9 +179,9 @@ void closeall(void)
   close(pidpipe[1]);
 }
 
-void callsplash(int n, char *path, char *action)
+void callsplash(int n, const char *path, char *action)
 {
-  char *p;
+  const char *p;
   char sbuf[32];
   char tbuf[256];
   pid_t pid;
@@ -248,11 +277,17 @@ static int checksystem(const int par, const boolean start, const boolean limit)
   if (read_proc(&prcs_run, &prcs_blked))
     return par;
 
+#ifdef USE_PRELOAD
+  /* if we have preload running, we expect I/O not to be a problem */
+  if (ispreload != NoPreload)
+    prcs_blked = 0;
+#endif
+
   newpar  = (par*numcpu) - prcs_run + 1; 	/* +1 for startpar its self */
   newpar -= (int)(((double)prcs_blked)*iorate);	/* I/O load reduction */
 
 #if DEBUG
-  fprintf(stderr, "checksystem par=%d newpar=%d (prcs_run=%u) %ld\n", par, newpar, prcs_run, time(0));
+  fprintf(stderr, "checksystem par=%d newpar=%d (prcs_run=%lu) %ld\n", par, newpar, prcs_run, time(0));
   dump_status();
 #endif
   if (newpar <= 0)
@@ -271,9 +306,75 @@ static inline int checkpar(const int par, const boolean start)
   return checksystem(par, start, false);
 }
 
+#ifdef CHECK_FORDEVPTS
+/*
+ * Based on __posix_openpt() from glibc.  Reimplemented here to work
+ * around the problem with getpt() failing for the entire process life
+ * time if /dev/pts/ is missing the first time it is called but
+ * mounted while the process is running.  BSD style pts is not
+ * supported, but might be copied from glibc too if there is need.
+ */
+# define DEVFS_SUPER_MAGIC       0x1373
+# define DEVPTS_SUPER_MAGIC      0x1cd1
+
+static int startpar_getpt(void) {
+  int fd = open("/dev/ptmx", O_RDWR|O_NOCTTY);
+
+  if (fd != -1)
+    {
+      struct statfs fsbuf;
+
+      /* Check that the /dev/pts filesystem is mounted
+        or if /dev is a devfs filesystem (this implies /dev/pts).  */
+      if ((statfs ("/dev/pts", &fsbuf) == 0
+             && fsbuf.f_type == DEVPTS_SUPER_MAGIC)
+         || (statfs ("/dev", &fsbuf) == 0
+             && fsbuf.f_type == DEVFS_SUPER_MAGIC))
+        {
+          /* Everything is ok, switch to the getpt() in libc.  */
+          return fd;
+        }
+
+      /* If /dev/pts is not mounted then the UNIX98 pseudo terminals
+        are not usable.  */
+      close (fd);
+    }
+
+  return -1;
+}
+
+static int checkdevpts(void)
+{
+  int ptsfd = startpar_getpt();
+
+  if (ptsfd == -1)
+    {
+      return 0;
+    }
+  else if (ptsname(ptsfd) == 0 || grantpt(ptsfd) || unlockpt(ptsfd))
+    {
+      close(ptsfd);
+      return 0;
+    }
+  else
+    {
+      close(ptsfd);
+      return 1;
+    }
+}
+#else
+static int __attribute__ ((unused)) checkdevpts(void)
+{
+  return 1;	/* /dev/pts is always mounted */
+}
+#endif
+
 void run(struct prg *p)
 {
-  char *m = 0;
+  const char * m = (char*)0;
+#ifdef USE_PRELOAD
+  pid_t parent;
+#endif
 
   p->len = 0;
   p->pid = (pid_t)0;
@@ -290,6 +391,9 @@ void run(struct prg *p)
       close(p->fd);
       p->fd = 0;
     }
+#ifdef USE_PRELOAD
+  parent = getpid();
+#endif
   if ((p->pid = fork()) == (pid_t)-1)
     {
       perror("fork");
@@ -341,21 +445,73 @@ void run(struct prg *p)
 
   closeall();
 
+#ifdef USE_PRELOAD
+  if (!strcmp(arg, "start")) 
+    { 
+      int s;
+
+      if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) != -1)
+	{
+	  struct sockaddr_un remote;
+	  pid_t child;
+	  int t, len;
+
+	  memset(&remote, 0, sizeof(struct sockaddr_un));
+	  remote.sun_family = AF_UNIX;
+	  strcpy(remote.sun_path, SOCK_PATH);
+	  len = strlen(remote.sun_path) + sizeof(remote.sun_family);
+
+	  if ((t = connect(s, (struct sockaddr *)&remote, len)) != -1)
+	    {
+	      char str[100];
+	      if (ispreload != Preload)
+		kill(parent, SIGUSR1);
+	      send(s, p->name, strlen(p->name), 0);
+	      recv(s, str, sizeof(str), 0);
+	    } 
+	  else if (ispreload == Unknown) 
+	    {
+	      /*
+	       * if we connected to preload once, we know it ran.
+	       * In case we can't connect to it later, it means it did
+	       * its job and we can guess I/O is no longer a problem. 
+	       */
+	      kill(parent, SIGUSR2);
+	    }
+	  close(s);
+	  /*
+	   * if we use preload, we fork again to make bootcharts easier to read.
+	   * The reason is that the name of the init script will otherwise be used
+	   * when in reality the above code waited for preload. If we fork away
+	   * before the exec, the waiting code will be folded into startpar
+	   */
+	  if ((child = fork()))
+	    {
+	      int status;
+	      int ret = waitpid(child, &status, 0);
+	      if (ret == -1)
+		perror("waitpid");
+	      exit(WEXITSTATUS(status));
+	    }
+	}
+    }
+#endif
+
   if (run_mode)
     {
       char path[128];
       snprintf(path, sizeof(path), "/etc/init.d/%s", p->name);
-      execlp(path, path, arg, (char *)0);
+      execlp(path, p->arg0, arg, (char *)0);
     }
   else if (arg)
-    execlp(p->name, p->name, arg, (char *)0);
+    execlp(p->name, p->arg0, arg, (char *)0);
   else
-    execlp(p->name, p->name, (char *)0);
+    execlp(p->name, p->arg0, (char *)0);
   perror(p->name);
   _exit(1);
 }
 
-int run_single(char *prg, int spl)
+int run_single(const char *prg, const char *arg0, int spl)
 {
   pid_t pid;
   int r;
@@ -382,12 +538,12 @@ int run_single(char *prg, int spl)
 	{
 	  char path[128];
 	  snprintf(path, sizeof(path), "/etc/init.d/%s", prg);
-	  execlp(path, path, arg, (char *)0);
+	  execlp(path, arg0 ? arg0 : path, arg, (char *)0);
 	}
       else if (arg)
-	execlp(prg, prg, arg, (char *)0);
+	execlp(prg, arg0 ? arg0 : prg, arg, (char *)0);
       else
-	execlp(prg, prg, (char *)0);
+	execlp(prg, arg0 ? arg0 : prg, (char *)0);
       perror(prg);
       _exit(1);
     }
@@ -415,6 +571,7 @@ void do_forward(void)
 	      rr = r;
 	    }
 	  r -= rr;
+	  b += rr;
 	}
     }
   _exit(0);
@@ -710,11 +867,11 @@ int main(int argc, char **argv)
 	{
 	  if ((*nodevec = pickup_task()))
 	  {
-	    *resvec = run_single((*nodevec)->name, calcsplash(0, 1, splashopt));
+	    *resvec = run_single((*nodevec)->name, (*nodevec)->arg0, calcsplash(0, 1, splashopt));
 	    finish_task(*nodevec);
 	  }
       } else
-	*resvec = run_single(*argv, calcsplash(0, 1, splashopt));
+	*resvec = run_single(*argv, *argv, calcsplash(0, 1, splashopt));
       goto finished;
     }
 
@@ -764,6 +921,26 @@ int main(int argc, char **argv)
       exit(1);
     }
 
+#ifdef USE_PRELOAD
+  sa.sa_handler = sighandler_preload;
+  sa.sa_flags = SA_RESTART;
+  (void)sigemptyset(&sa.sa_mask);
+  if (sigaction(SIGUSR1, &sa, 0))
+    {
+      perror("sigusr1(preload) sigaction");
+      exit(1);
+    }
+
+  sa.sa_handler = sighandler_nopreload;
+  sa.sa_flags = SA_RESTART;
+  (void)sigemptyset(&sa.sa_mask);
+  if (sigaction(SIGUSR2, &sa, 0))
+    {
+      perror("sigusr2(preload) sigaction");
+      exit(1);
+    }
+#endif
+
   gettimeofday(&glastio, 0);
   limit = checklimit(inpar, isstart);
   lastlim = glastio;
@@ -775,6 +952,9 @@ int main(int argc, char **argv)
       pid_t pid = 0;
       int r = 0, s;
       long diff;
+#ifdef CHECK_FORDEVPTS
+      int devpts = 0;
+#endif
 
       gettimeofday(&now, 0);
       FD_ZERO(&rset);
@@ -796,6 +976,10 @@ int main(int argc, char **argv)
       for (s = 0; s < par; s++)			/* never leave this with break!! */
 	{
 	account:				/* for the new process below */
+#ifdef CHECK_FORDEVPTS
+	  if (!devpts)
+	    devpts = checkdevpts();
+#endif
 	  p = prgs + s;
 	  if (p == interactive_task)
 	    continue;				/* don't count this here */
@@ -818,13 +1002,23 @@ int main(int argc, char **argv)
 		      if (nodevec[num]->interactive)
 			interactive_task = p;
 		      p->name = nodevec[num]->name;
+		      p->arg0 = nodevec[num]->arg0 ? nodevec[num]->arg0 : nodevec[num]->name;
 		    }
-		  else
+		  else {
 		    p->name = *argv++;
+		    p->arg0 = p->name;
+		  }
 		  p->splashadd = calcsplash(num, argc, splashopt);
 		  p->num = num++;
 		  if (interactive_task)
 		    continue;			/* don't start this here */
+#ifdef CHECK_FORDEVPTS
+		  if (!devpts)
+		    {
+		      interactive_task = p;	/* no /dev/pts, treat as interactive */
+		      continue;
+		    }
+#endif
 		  run(p);
 		  if (p->pid == 0)
 		    {
@@ -855,7 +1049,7 @@ int main(int argc, char **argv)
 	  if (active == 0)
 	    {
 	      p = interactive_task;
-	      resvec[p->num] = run_single(p->name, p->splashadd);
+	      resvec[p->num] = run_single(p->name, p->arg0, p->splashadd);
 	      if (run_mode)
 		finish_task(nodevec[p->num]);
 	      p->pid = 0;
